"""Functions for calculating similarity metrics on preprocessed data."""

from typing import Optional

import numpy as np

from mobi_motion_tracking.core import models


def dynamic_time_warping(
    preprocessed_target_data: np.ndarray,
    preprocessed_subject_data: np.ndarray,
    window_size: Optional[int] = None,
) -> models.SimilarityMetrics:
    """Perform dynamic time warping.

    This function calculates the cumulative distance and optimal warping paths between
    the subject and target sequences. Both sequences have been centered to the
    hip and the experimental sequence has been normalized to the target lengths.
    Both input sequences have 61 columns. Column 0 contains frame number. Columns 1-3
    contain the x, y, and z coordinates of the hip. Since the data has already been
    centered to the hip, when calculating the distance and paths between the sequences,
    we index the data starting at the 4th column until the end. This function returns
    a dataclass which stores the DTW similarity metrics, distance and paths.

    Args:
        preprocessed_target_data: cleaned and centered target data.
        preprocessed_subject_data: cleaned, centered, and normalized subject data.
        window_size: constraint for matching points, ensuring
            |num_frames_subject - num_frames_target| <= window_size. If None, the
            window size is set to the maximum amount of rows of the two sequences.

    Returns:
        SimilarityMetrics: a dataclass which stores the DTW similarity metrics.

    Raises:
        ValueError: when dimensions of the two inputs do not match.
    """
    preprocessed_subject_data = preprocessed_subject_data[:, 4:]
    preprocessed_target_data = preprocessed_target_data[:, 4:]

    num_frames_subject, num_joints_subject = preprocessed_subject_data.shape
    num_frames_target, num_joints_target = preprocessed_target_data.shape

    if num_joints_subject != num_joints_target:
        raise ValueError(
            "Error in dtw(): the dimensions of the two input signals do not match."
        )

    if window_size is None:
        window_size = max(num_frames_subject, num_frames_target)
    else:
        window_size = max(window_size, abs(num_frames_subject - num_frames_target))

    cost_matrix = np.full((num_frames_subject + 1, num_frames_target + 1), float("inf"))
    cost_matrix[0, 0] = 0

    for row in range(1, num_frames_subject + 1):
        for column in range(
            max(1, row - window_size), min(num_frames_target + 1, row + window_size + 1)
        ):
            cost_value = np.linalg.norm(
                preprocessed_subject_data[row - 1]
                - preprocessed_target_data[column - 1]
            )
            cost_matrix[row, column] = cost_value + min(
                cost_matrix[row - 1, column],
                cost_matrix[row, column - 1],
                cost_matrix[row - 1, column - 1],
            )

    distance = cost_matrix[num_frames_subject, num_frames_target]

    subject_idx, target_idx = num_frames_subject, num_frames_target
    path = [(subject_idx, target_idx)]

    while subject_idx > 0 or target_idx > 0:
        if subject_idx == 0:
            target_idx -= 1
        elif target_idx == 0:
            subject_idx -= 1
        else:
            min_cost_index = np.argmin(
                [
                    cost_matrix[subject_idx - 1, target_idx],
                    cost_matrix[subject_idx, target_idx - 1],
                    cost_matrix[subject_idx - 1, target_idx - 1],
                ]
            )
            if min_cost_index == 0:
                subject_idx -= 1
            elif min_cost_index == 1:
                target_idx -= 1
            else:
                subject_idx -= 1
                target_idx -= 1
        path.append((subject_idx, target_idx))

    path.reverse()

    return models.SimilarityMetrics.from_dtw(distance=distance, warping_path=path)



"""Functions to read motion tracking data from a file."""

import pathlib

import numpy as np
import pandas as pd

from mobi_motion_tracking.core import models


def data_cleaner(data: pd.DataFrame) -> np.ndarray:
    """Select applicable data from a dataframe.

    Currently motion tracking data for Kinect and Zed are saved into xlsx files.
    This function searches through dataframe extracted from the read_sheet function for
    x_hip, then extracts all rows below and the neighboring 61 columns.

    Args:
        data: dataframe from read_sheet.

    Returns:
        cleaned_data: np.array of [x,61] where x is the total number of rows
            representing all frames and columns representing the frame number and 60
            joint coordinates.

    Raises:
        ValueError: when x_Hip is not found in dataframe.
        IndexError: when column index is out of range.
    """
    result = data.where(data == "x_Hip").stack().index

    if result.empty:
        raise ValueError("x_Hip not found in DataFrame.")

    row = result[0][0]
    col_idx = result[0][1]

    start_col = data.columns.get_loc(col_idx) - 1
    end_col = data.columns.get_loc(col_idx) + 60

    if start_col < 0 or end_col > data.shape[1]:
        raise IndexError("Column index out of range.")

    cleaned_data = (
        data.iloc[
            row + 1 :,
            start_col:end_col,
        ]
        .to_numpy()
        .astype(np.float64)
    )

    return cleaned_data


def read_participant_data(
    subject_path: pathlib.Path, sequence: int
) -> models.ParticipantData:
    """Calls get_metadata and read sheet.

    This function calls get_metadata to extract the participant_ID value and the
    sequence_sheetname. read_sheet is then called to create the subject_data array.
    The participant_ID, sequence_sheetname, and subject_data are passed to the
    ParticipantData class and returned.

    Args:
        subject_path: file path to the participant file.
        sequence: integer value indiciating the sequence currently being tested.

    Returns:
        models.ParticipantData: containing participant_ID (str), sheetname (str),
            and data (np.ndarray).
    """
    participant_ID = subject_path.stem
    sequence_sheetname = f"seq{sequence}"

    try:
        motion_tracking_data = pd.read_excel(
            subject_path, sheet_name=sequence_sheetname, engine="openpyxl"
        )
    except ValueError as ve:
        print(f"Sheet doesn't exist: {ve}")
        return models.ParticipantData(
            participant_ID=participant_ID,
            sequence_sheetname=sequence_sheetname,
            data=np.array([]),
        )

    subject_data = data_cleaner(motion_tracking_data)

    return models.ParticipantData(
        participant_ID=participant_ID,
        sequence_sheetname=sequence_sheetname,
        data=subject_data,
    )



"""The core submodule contains key classes and functions.

These classes and functions are used to organize the pipeline processing
and provide functions for data organization and output analysis.
"""



"""Dataclass storing all similarity metrics."""

from dataclasses import dataclass, field
from typing import Any, Dict

import numpy as np


@dataclass
class SimilarityMetrics:
    """Stores similarity metrics between two time-series sequences.

    Attributes:
        method: A string containing the name of the similarity function used.
        metrics A list containing speciifc outputs for each
            similarity function.
    """

    method: str
    metrics: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dtw(
        cls, distance: float, warping_path: list[tuple[int, int]]
    ) -> "SimilarityMetrics":
        """Creates a SimilarityMetrics instance from DTW output.

        Args:
            distance: The cumulative distance between experimental and target sequences.
            warping_path: A list of tuples representing the warping path.

        Returns:
            A SimilarityMetrics instance storing DTW-specific metrics.
        """
        return cls(
            method="DTW",
            metrics={
                "distance": distance,
                "target_path": ([p[0] for p in warping_path]),
                "experimental_path": ([p[1] for p in warping_path]),
            },
        )


@dataclass
class ParticipantData:
    """Stores relevant participant information.

    Attributes:
        participant_ID: Integer value unique to every participant.
        sequence: Integer value associated with each iteration of a test.
        data: np.ndarray containing subject data for specified sequence.
    """

    participant_ID: str
    sequence_sheetname: str
    data: np.ndarray
